@@ -1,16 +1,27 @@
 import os
 import re
-import time
 import random
-import socket
 import string
 import subprocess
+import warnings
+import logging
 
 from tempfile import mkdtemp
 from datetime import datetime
-from urllib import urlencode
+from functools import reduce
 
-from hetzner import RobotError, ManualReboot, ConnectError
+try:
+    from urllib import urlencode
+except ImportError:
+    from urllib.parse import urlencode
+
+from hetzner import RobotError, WebRobotError
+from hetzner.rdns import ReverseDNS, ReverseDNSManager
+from hetzner.reset import Reset
+from hetzner.util import addr, scraping
+
+__all__ = ['AdminAccount', 'IpAddress', 'RescueSystem', 'Server', 'Subnet',
+           'IpManager', 'SubnetManager']
 
 
 class SSHAskPassHelper(object):
@@ -26,10 +37,11 @@ class SSHAskPassHelper(object):
     def __enter__(self):
         self.tempdir = mkdtemp()
         script = os.path.join(self.tempdir, "askpass")
-        fd = os.open(script, os.O_WRONLY | os.O_CREAT | os.O_NOFOLLOW, 0700)
+        fd = os.open(script, os.O_WRONLY | os.O_CREAT | os.O_NOFOLLOW, 0o700)
         self.script = script
         esc_passwd = self.passwd.replace("'", r"'\''")
-        os.write(fd, "#!/bin/sh\necho -n '{0}'".format(esc_passwd))
+        askpass = "#!/bin/sh\necho -n '{0}'".format(esc_passwd).encode('ascii')
+        os.write(fd, askpass)
         os.close(fd)
         return script
 
@@ -117,6 +129,11 @@ class RescueSystem(object):
 
         Look at Server.observed_reboot() for further options.
         """
+        msg = ("The RescueSystem.shell() method will be removed from the API"
+               " in version 1.0.0, please do not use it! See"
+               " https://github.com/aszlig/hetzner/issues/13"
+               " for details.")
+        warnings.warn(msg, FutureWarning)
         self.observed_activate(*args, **kwargs)
 
         with SSHAskPassHelper(self.password) as askpass:
@@ -160,7 +177,7 @@ class AdminAccount(object):
         path = '/server/admin/id/{0}'.format(self._serverid)
         response = self._scraper.request(path)
         assert response.status == 200
-        match = login_re.search(response.read())
+        match = login_re.search(response.read().decode('utf-8'))
         if match is None:
             self.exists = False
         else:
@@ -169,7 +186,7 @@ class AdminAccount(object):
 
     def _genpasswd(self):
         random.seed(os.urandom(512))
-        chars = string.letters + string.digits + "/()-=+_,;.^~#*@"
+        chars = string.ascii_letters + string.digits + "/()-=+_,;.^~#*@"
         length = random.randint(20, 40)
         return ''.join(random.choice(chars) for i in range(length))
 
@@ -178,19 +195,44 @@ class AdminAccount(object):
         Create a new admin account if missing. If passwd is supplied, use it
         instead of generating a random one.
         """
-        if not self.exists:
-            path = '/server/adminCreate/id/{0}'.format(self._serverid)
-            self._scraper.request(path)
-            self.update_info()
         if passwd is None:
             passwd = self._genpasswd()
+
+        form_path = '/server/admin/id/{0}'.format(self._serverid)
+        form_response = self._scraper.request(form_path, method='POST')
+
+        parser = scraping.CSRFParser('password[_csrf_token]')
+        parser.feed(form_response.read().decode('utf-8'))
+        assert parser.csrf_token is not None
+
         data = {
             'password[new_password]': passwd,
             'password[new_password_repeat]': passwd,
-            'id': self._serverid
+            'password[_csrf_token]': parser.csrf_token,
         }
-        response = self._scraper.request('/server/adminUpdate', data)
-        assert "msgbox_success" in response.read()
+
+        if not self.exists:
+            failmsg = "Unable to create admin account"
+            path = '/server/adminCreate/id/{0}'.format(self._serverid)
+        else:
+            failmsg = "Unable to update admin account password"
+            path = '/server/adminUpdate'
+            data['id'] = self._serverid
+
+        response = self._scraper.request(path, data)
+        data = response.read().decode('utf-8')
+        if "msgbox_success" not in data:
+            ul_re = re.compile(r'<ul\s+class="error_list">(.*?)</ul>',
+                               re.DOTALL)
+            li_re = re.compile(r'<li>\s*([^<]*?)\s*</li>')
+            ul_match = ul_re.search(data)
+            if ul_match is not None:
+                errors = [error.group(1)
+                          for error in li_re.finditer(ul_match.group(0))]
+                msg = failmsg + ': ' + ', '.join(errors)
+                raise WebRobotError(msg)
+            raise WebRobotError(failmsg)
+        self.update_info()
         self.passwd = passwd
         return self.login, self.passwd
 
@@ -201,7 +243,8 @@ class AdminAccount(object):
         if not self.exists:
             return
         path = '/server/adminDelete/id/{0}'.format(self._serverid)
-        assert "msgbox_success" in self._scraper.request(path).read()
+        assert "msgbox_success" in \
+            self._scraper.request(path).read().decode('utf-8')
         self.update_info()
 
     def __repr__(self):
@@ -212,9 +255,20 @@ class AdminAccount(object):
 
 
 class IpAddress(object):
-    def __init__(self, conn, result):
+    def __init__(self, conn, result, subnet_ip=None):
         self.conn = conn
+        self.subnet_ip = subnet_ip
         self.update_info(result)
+        self._rdns = None
+
+    @property
+    def rdns(self):
+        """
+        Get or set reverse DNS PTRs.
+        """
+        if self._rdns is None:
+            self._rdns = ReverseDNS(self.conn, self.ip)
+        return self._rdns
 
     def update_info(self, result=None):
         """
@@ -222,10 +276,18 @@ class IpAddress(object):
         information such as traffic warnings. If result is omitted, a new
         request is sent to the robot to gather the information.
         """
-        if result is None:
-            result = self.conn.get('/ip/{0}'.format(self.ip))
-
-        data = result['ip']
+        if self.subnet_ip is not None:
+            if result is None:
+                result = self.conn.get('/subnet/{0}'.format(self._subnet_addr))
+            data = result['subnet']
+            self._subnet_addr = data['ip']
+            data['ip'] = self.subnet_ip
+            # Does not exist in subnets
+            data['separate_mac'] = None
+        else:
+            if result is None:
+                result = self.conn.get('/ip/{0}'.format(self.ip))
+            data = result['ip']
 
         self.ip = data['ip']
         self.server_ip = data['server_ip']
@@ -268,7 +330,7 @@ class Subnet(object):
         request is sent to the robot to gather the information.
         """
         if result is None:
-            result = self.conn.get('/ip/{0}'.format(self.net_ip))
+            result = self.conn.get('/subnet/{0}'.format(self.net_ip))
 
         data = result['subnet']
 
@@ -283,6 +345,37 @@ class Subnet(object):
         self.traffic_daily = data['traffic_daily']
         self.traffic_monthly = data['traffic_monthly']
 
+        self.is_ipv6, self.numeric_net_ip = addr.parse_ipaddr(self.net_ip)
+        self.numeric_gateway = addr.parse_ipaddr(self.gateway, self.is_ipv6)
+        getrange = addr.get_ipv6_range if self.is_ipv6 else addr.get_ipv4_range
+        self.numeric_range = getrange(self.numeric_net_ip, self.mask)
+
+    def get_ip_range(self):
+        """
+        Return the smallest and biggest possible IP address of the current
+        subnet.
+        """
+        convert = addr.ipv6_bin2addr if self.is_ipv6 else addr.ipv4_bin2addr
+        return convert(self.numeric_range[0]), convert(self.numeric_range[1])
+
+    def __contains__(self, addr):
+        """
+        Check whether a specific IP address is within the current subnet.
+        """
+        numeric_addr = addr.parse_ipaddr(addr, self.is_ipv6)
+        return self.numeric_range[0] <= numeric_addr <= self.numeric_range[1]
+
+    def get_ip(self, addr):
+        """
+        Return an IpAddress object for the specified IPv4 or IPv6 address or
+        None if the IP address doesn't exist in the current subnet.
+        """
+        if addr in self:
+            result = self.conn.get('/subnet/{0}'.format(self.net_ip))
+            return IpAddress(self.conn, result, addr)
+        else:
+            return None
+
     def __repr__(self):
         return "<Subnet {0}/{1} (Gateway: {2})>".format(self.net_ip, self.mask,
                                                         self.gateway)
@@ -316,9 +409,12 @@ class Server(object):
         self.conn = conn
         self.update_info(result)
         self.rescue = RescueSystem(self)
+        self.reset = Reset(self)
         self.ips = IpManager(self.conn, self.ip)
         self.subnets = SubnetManager(self.conn, self.ip)
+        self.rdns = ReverseDNSManager(self.conn, self.ip)
         self._admin_account = None
+        self.logger = logging.getLogger("Server #{0}".format(self.number))
 
     @property
     def admin(self):
@@ -345,107 +441,27 @@ class Server(object):
         self.product = data['product']
         self.datacenter = data['dc']
         self.traffic = data['traffic']
-        self.flatrate = data['flatrate']
         self.status = data['status']
-        self.throttled = data['throttled']
         self.cancelled = data['cancelled']
         self.paid_until = datetime.strptime(data['paid_until'], '%Y-%m-%d')
         self.is_vserver = self.product.startswith('VQ')
 
+    def observed_reboot(self, *args, **kwargs):
+        msg = ("Server.observed_reboot() is deprecated. Please use"
+               " Server.reset.observed_reboot() instead.")
+        warnings.warn(msg, DeprecationWarning)
+        return self.reset.observed_reboot(*args, **kwargs)
+
+    def reboot(self, *args, **kwargs):
+        msg = ("Server.reboot() is deprecated. Please use"
+               " Server.reset.reboot() instead.")
+        warnings.warn(msg, DeprecationWarning)
+        return self.reset.reboot(*args, **kwargs)
+
     def set_name(self, name):
         result = self.conn.post('/server/{0}'.format(self.ip),
                                 {'server_name': name})
         self.update_info(result)
 
-    def check_ssh(self, port=22, timeout=5):
-        """
-        Check if the current server has an open SSH port. Return True if port
-        is reachable, otherwise false. Time out after 'timeout' seconds.
-        """
-        success = True
-        old_timeout = socket.getdefaulttimeout()
-        socket.setdefaulttimeout(5)
-
-        try:
-            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-            s.connect((self.ip, port))
-            s.close()
-        except socket.error:
-            success = False
-
-        socket.setdefaulttimeout(old_timeout)
-        return success
-
-    def observed_reboot(self, patience=300, tries=None, manual=False):
-        """
-        Reboot and wait patience seconds until the system comes back.
-        If not, retry with the next step in tries and wait another patience
-        seconds. Repeat until there are no more tries left.
-
-        If manual is true, do a manual reboot in case the server doesn't come
-        up again. Raises a ManualReboot exception if that is the case.
-
-        Return True on success and False if the system didn't come up.
-        """
-        is_down = False
-
-        if tries is None:
-            if self.is_vserver:
-                tries = ['hard']
-            else:
-                tries = ['soft', 'hard']
-
-        for mode in tries:
-            self.reboot(mode)
-
-            now = time.time()
-            while True:
-                if time.time() > now + patience:
-                    break
-
-                is_up = self.check_ssh()
-                time.sleep(1)
-
-                if is_up and is_down:
-                    return
-                elif not is_down:
-                    is_down = not is_up
-        if manual:
-            self.reboot('manual')
-            raise ManualReboot("Issued a manual reboot because the server"
-                               " did not come back to life.")
-        else:
-            raise ConnectError("Server keeps playing dead after reboot :-(")
-
-    def reboot(self, mode='soft'):
-        """
-        Reboot the server, modes are "soft" for reboot by triggering Ctrl-Alt-
-        Del, "hard" for triggering a hardware reset and "manual" for requesting
-        a poor devil from the data center to go to your server and press the
-        power button.
-
-        On a vServer, rebooting with mode="soft" is a no-op, any other value
-        results in a hard reset.
-        """
-        if self.is_vserver:
-            if mode == 'soft':
-                return
-
-            self.conn.scraper.login(force=True)
-            baseurl = '/server/vserverCommand/id/{0}/command/reset'
-            url = baseurl.format(self.number)
-            response = self.conn.scraper.request(url, method='POST')
-            assert "msgbox_success" in response.read()
-            return response
-
-        modes = {
-            'manual': 'man',
-            'hard': 'hw',
-            'soft': 'sw',
-        }
-
-        modekey = modes.get(mode, modes['soft'])
-        return self.conn.post('/reset/{0}'.format(self.ip), {'type': modekey})
-
     def __repr__(self):
         return "<{0} (#{1} {2})>".format(self.ip, self.number, self.product)
